
PROGRAM MILLSR2M3

    !!***********************************************************************
    !!
    !!.........................................................................
    !!  program body      starts at line   77
    !!  subroutine MAKEGRD starts at line 343
    !!  subroutine MAKEBDY starts at line 405
    !!
    !!  DESCRIPTION:
    !!       Builds 1-layer/1-variable gridded file with LAT and LON
    !!
    !!  PRECONDITIONS REQUIRED:
    !!
    !!  REVISION  HISTORY:
    !!      prototype 06/2018 by Carlie J. Coats, Jr.
    !!***********************************************************************

    USE M3UTILIO

    IMPLICIT NONE

    !!...........   PARAMETERS and their descriptions:

    CHARACTER*16, PARAMETER :: PNAME = 'MILLSR2M3'
    CHARACTER*16, PARAMETER :: BLANK = ' '
    CHARACTER*72, PARAMETER :: BAR   = &
    '-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-'

    CHARACTER*20, PARAMETER :: CMENU( 10 ) = &
        (/  'lat-lon (ISO 6709)   ',        &   !  coordinate types menu item  1
            'Lambert Conformal    ',        &   !  coordinate types menu item  2
            'Lambert Equal Area   ',        &   !  coordinate types menu item  3
            'Equatorial Mercator  ',        &   !  coordinate types menu item  4
            'Transverse Mercator  ',        &   !  coordinate types menu item  5
            'General    Mercator  ',        &   !  coordinate types menu item  6
            'Polar  Stereographic ',        &   !  coordinate types menu item  7
            'General Stereographic',        &   !  coordinate types menu item  8
            'UTM                  ',        &   !  coordinate types menu item  9
            'Albers Equal-Area    '  /)         !  coordinate types menu item 10

    INTEGER, PARAMETER :: CTYPE( 10 ) =  &
        (/ LATGRD3, LAMGRD3, LEQGRD3, EQMGRD3, TRMGRD3, MERGRD3, POLGRD3, STEGRD3, UTMGRD3, ALBGRD3 /)

    !!...........   LOCAL VARIABLES and their descriptions:

    INTEGER         LOGDEV, ISTAT, IREC
    INTEGER         NCOLS, NROWS, NGRID, NRECS
    INTEGER         SDATE, STIME, TSTEP, JDATE, JTIME, EDATE, ETIME
    INTEGER         FDEV
    LOGICAL         EFLAG
    CHARACTER*16    ANAME, BNAME, GNAME
    CHARACTER*512   MESG
    CHARACTER*512   PATH

    REAL     , ALLOCATABLE :: RBUF( :,: )


    !!***********************************************************************
    !!.......   First:  Initialize the I/O API:

    EFLAG = .FALSE.
    LOGDEV = INIT3()	!  initialization returns unit # for log

    WRITE( *, '( 5X, A )' ) BLANK, BAR, BLANK,                              &
'Program MILLSR2M3 to convert an LSHM gridded time stepped single-variable',   &
'REAL direct access (binary) input file to an equivalent M3IO file.',       &
'',                                                                         &
'PRECONDITIONS REQUIRED:',                                                  &
'    setenv INFILE     <path name for  input time stepped',                 &
'                       GRIDFLOAT gridded file>',                           &
'    setenv M3FILE     <path name for output file.',                        &
'',                                                                         &
'Default dimensions are for the Mills River LSHM domain.',                  &
'',                                                                         &
'THE PROGRAM WILL PROMPT YOU for starting and ending date&time and',        &
'time step for the data, the name, description, and dimensions for',        &
'the input variable.',                                                      &
'',                                                                         &
'Copyright (C) 2018 Carlie J. Coats, Jr.',                                  &
' '

    IF ( .NOT.GETYN( 'Continue with program?', .TRUE. ) ) THEN
        CALL M3EXIT( PNAME, 0, 0, 'Program aborted', 2 )
    END IF

    CALL M3MESG( BAR )
    CALL M3MESG( 'First, let us get the dimensions and (partial) file description.' )

    NCOLS3D = GETVAL( 1, 1999999999, 112, 'Enter number of GRID COLUMNS' )
    NROWS3D = GETVAL( 1, 1999999999,  76, 'Enter number of GRID ROWS' )
    NLAYS3D = 1
    NTHIK3D = 1
    XCELL3D = GETVAL( 250.0d0, 'Enter X-cell size' )
    YCELL3D = GETVAL( 250.0d0, 'Enter Y-cell size' )
    P_ALP3D = 0.0d0
    P_BET3D = 0.0d0
    P_GAM3D = 0.0d0
    XCENT3D = 0.0d0
    YCENT3D = 0.0d0
    XORIG3D = 0.0d0
    YORIG3D = 0.0d0
    FTYPE3D = GRDDED3
    SDATE3D = GETVAL( 2007001,          'Enter initial date (YYYYDDD)' )
    STIME3D = GETVAL( 0, 99999999, 0,   'Enter initial time  (HHMMSS)' )
    TSTEP3D = GETVAL( 0, 99999999, 500, 'Enter time step     (HHMMSS)' )
    SDATE   = SDATE3D
    STIME   = STIME3D
    JDATE   = SDATE3D
    JTIME   = STIME3D
    TSTEP   = TSTEP3D
    CALL NEXTIME( JDATE, JTIME, 30*240000 )
    EDATE = GETVAL( SDATE, 9999999, JDATE, 'Enter  final  date (YYYYDDD)' )
    ETIME = GETVAL( 0, 99999999, 235500,   'Enter  final  time  (HHMMSS)' )
    NRECS = CURREC( EDATE, ETIME, SDATE, STIME, TSTEP3D, JDATE, JTIME )
    NCOLS = NCOLS3D
    NROWS = NROWS3D

    NVARS3D = 1
    CALL GETSTR( 'Enter name        for variable', 'FOO',  VNAME3D(1) )
    CALL GETSTR( 'Enter units       for variable', 'Bars', UNITS3D(1) )
    CALL GETSTR( 'Enter description for variable', 'qux',  VDESC3D(1) )
    VTYPE3D(1) = M3REAL

    FDESC3D    = BLANK
    FDESC3D(1) = 'Variable transposed/copied from LSHM gridded file'

    IF ( .NOT.OPEN3( 'M3FILE', FSUNKN3, PNAME ) ) THEN
        CALL M3EXIT( PNAME, 0,0, 'Could not open "M3FILE"' , 2 )
    END IF

    CALL NAMEVAL( 'INFILE', PATH )
    FDEV = JUNIT()
    OPEN( FDEV, FILE   = TRIM( PATH ),     &
                RECL   = 4*NCOLS*NROWS,    &
                ACTION = 'read',           &
                ACCESS = 'direct',         &
                FORM   = 'unformatted',    &
                STATUS = 'old',            &
                IOSTAT = ISTAT )
    IF ( ISTAT .NE. 0 ) THEN
        MESG = 'INFILE="' // TRIM( PATH ) // '"'
        CALL M3MESG( MESG )
        WRITE( MESG, '( A, I9 )' )  'ERROR opening "INFILE": STATUS=', ISTAT
        CALL M3EXIT( PNAME, 0, 0, MESG, 2 )
    END IF

    ALLOCATE( RBUF( NCOLS,NROWS ), STAT = ISTAT )
    IF ( ISTAT .NE. 0 ) THEN
        WRITE( MESG, '( A, I10 )' ) 'ERROR:  Allocation failure -- STATUS=', ISTAT
        CALL M3EXIT( PNAME, 0, 0, MESG, 2 )
    END IF

    JDATE = SDATE3D
    JTIME = STIME3D

    DO IREC = 1, NRECS

        IF ( .NOT.RDXPREC( FDEV, IREC, NCOLS, NROWS, RBUF ) ) THEN
            EFLAG = .TRUE.
        ELSE IF ( .NOT.WRITE3( 'M3FILE', VNAME3D(1), JDATE, JTIME, RBUF ) ) THEN
            EFLAG = .TRUE.
        END IF

        CALL NEXTIME( JDATE, JTIME, TSTEP )

    END DO


    IF ( EFLAG ) THEN
        MESG  = 'Failure(s) in program'
        ISTAT = 2
    ELSE
        MESG  = 'Success in program'
        ISTAT = 0
    END IF

    CALL M3EXIT( PNAME, 0, 0, MESG, ISTAT )


CONTAINS    !!-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-


    LOGICAL FUNCTION RDXPREC( FDEV, IREC, NCOLS, NROWS, RBUF )
        INTEGER, INTENT(IN   ) :: FDEV, IREC, NCOLS, NROWS
        REAL   , INTENT(  OUT) :: RBUF( NCOLS, NROWS )

        REAL            XBUF( NCOLS ), T
        INTEGER         C, R, K, ISTAT
        CHARACTER*80    MESG

        READ( FDEV, REC=IREC, IOSTAT=ISTAT ) RBUF
        IF ( ISTAT .NE. 0 ) THEN
            WRITE( MESG, '( A, I10, 2X, A, I10)' )      &
                'RDXPREC:  record', IREC, 'failure -- ERROR-STATUS=', ISTAT
            CALL M3MESG( MESG )
            RDXPREC = .FALSE.
            RETURN
        END IF

        DO R = 1, NROWS/2       !! transpose RBUF:
            K = NROWS - R + 1
            DO C = 1, NCOLS
                T           = RBUF( C,K )
                RBUF( C,K ) = RBUF( C,R )
                RBUF( C,R ) = T
            END DO
        END DO

        RDXPREC = .TRUE.
        RETURN

    END FUNCTION RDXPREC

END PROGRAM MILLSR2M3
