
C Version "%W% %P% %G% %U%

        SUBROUTINE  MMD2P( IM, JM, KM, 
     &                     PP, PST, SIGH, PTOP, PLEV,
     &                     NCOLS, NROWS, NLAYS,
     &                     COL0, ROW0, LAY0, 
     &                     LEVS, CLIN, CLOG, CPOW )

C***********************************************************************
C  subroutine body starts at line  88
C
C  FUNCTION:  Extract/reorder/convert/decouple/compute the dot point
C	linear, log, and power pressure-surface interpolation coefficients 
C       for the window specified in terms of grid-window bounds to I/O API 
C       order, using Pprime, Pstar, SIGH, and Ptop from MM5. The
C       interpolation coefficients allow conversion from sigma to pressure
C       coordinates.
C
C  NOTE that MM5-order I subscripts go from south to north, J subscripts
C       go from west to east, and K subscripts go from top to bottom,
C       whereas IOAPI-order C subscripts go from west to east,
C       R subscripts go from south to north, and L subscripts go
C       from bottom to top.
C
C  PRECONDITIONS REQUIRED:
C       ioapi grid fits within the MM5 grid
C
C  SUBROUTINES AND FUNCTIONS CALLED: none
C
C  REVISION  HISTORY:
C       Feb. 09, 1999 by Carlie J. Coats, Jr. & John N. McHenry, MCNC
C       Validated 2/16/99 CJC, JNM.
C             Note: this version produces interpolation coefficients
C                   that effectively extrapolate for pressure surfaces
C                   that are "underground" due to terrain. An example
C                   would be the 95000Pa surface in the Rockies.
C
C***********************************************************************

      IMPLICIT NONE

C...........   ARGUMENTS and their descriptions:

        INTEGER         IM      !  I (north-south) dimension for AM
        INTEGER         JM      !  J (east-west)   dimension for AM
        INTEGER         KM      !  K (top-bottom)  dimension for AM
        REAL            PP( IM, JM, KM )        !  MM5 coupled P-prime (CB-Pa)
        REAL            PST( IM, JM )           !  MM5 pstar array (CB)
        REAL            SIGH( KM )              !  MM5 SIGMA-H
        REAL            PTOP			!  MM5 ptop (CB)
        REAL            PLEV( NLAYS )           !  requested press levels (Pa)
        INTEGER         NCOLS   !  output dimension
        INTEGER         NROWS   !  " 
        INTEGER         NLAYS   !  "
        INTEGER         COL0    !  starting subscript for output window
        INTEGER         ROW0    !  "
        INTEGER         LAY0    !  "
        INTEGER         LEVS( NCOLS, NROWS, NLAYS ) !  level at or below PLEV
        REAL            CLIN( NCOLS, NROWS, NLAYS ) !  coeff -- lin interp
        REAL            CLOG( NCOLS, NROWS, NLAYS ) !  coeff -- log interp
        REAL            CPOW( NCOLS, NROWS, NLAYS ) !  coeff -- power interp

C...........   PARAMETERs

        INTEGER         MAXK             !  dimension for scratch arrays
        REAL	    	CBTOPA           !  CB to Pa
        REAL	    	RDGAS		 !  from param.F, line 1573 & ff
	REAL            CP		 !  "
	REAL            ROCP		 !  for p**kappa interpolation

        PARAMETER     ( CBTOPA = 1.0E3 ,
     &                  MAXK   = 200 ,
     &                  RDGAS =  287.04, !  to preserve MM5 compatibility
     &                  CP    = 1004.0 , !  "
     &                  ROCP  = RDGAS / CP  )   !  INTERP "kappa"
	       	

C...........   SCRATCH LOCAL VARIABLES and their descriptions:

	INTEGER		C, R, L		        !  IOAPI col subscript
        INTEGER         I, J, K                 !  MM5 subscripts
        INTEGER         M, N                    !  loop counters
        INTEGER         I0, J0, K0              !  MM5 offsets
        
        REAL            DPP		        !  decoupled PP
        REAL            PRES( MAXK )
        REAL            PLOG( MAXK )
        REAL            PPOW( MAXK )

C***********************************************************************
C   begin body of subroutine  MMD2P

        IF ( KM .GT. MAXK ) THEN
            CALL M3EXIT( 'MCPL:MMD2P', 0, 0, 
     &      'MAXK too small; increase and recompile', 2 )
        END IF

        I0 = ROW0 - 1                   !  offsets for IOAPI grid within MM5
        J0 = COL0 - 1
        K0 = KM - LAY0 + 2

        DO  33  R = 1, NROWS
            I = I0 + R
            DO  22  C = 1, NCOLS
                J = J0 + C
                
                DO  L = 1, KM   
                
                !  loop computing SIG-level dot-pt pressures in I/O API order:
                !  Compute 4-point mean of decoupled cross-pt PPRIME,
                !  4-point mean of cross-pt PSTAR 
                !  Use these to compute pressure (Pa) and its log and power

                    K = K0 - L
                    DPP  = 0.25 * ( PP( I-1,J-1,K )/PST( I-1,J-1 ) +
     &                              PP( I  ,J-1,K )/PST( I  ,J-1 ) +
     &                              PP( I-1,J  ,K )/PST( I-1,J   ) +
     &                              PP( I  ,J  ,K )/PST( I  ,J   ) ) !  ", pa
                    PRES(L) = CBTOPA*( SIGH(K)*0.25*( PST( I-1,J-1 ) +
     &                                                PST( I  ,J-1 ) +
     &                                                PST( I-1,J   ) +
     &                                                PST( I  ,J   ) )
     &                               + PTOP ) + DPP
                    PLOG(L) = LOG( PRES(L) )
                    PPOW(L) = PRES(L)**ROCP
                END DO
                
                N = 1
                DO  L = 1, NLAYS        !  loop on output layers

                    DO  M = N+1, KM !  search remaining MM5 levels

                        IF ( PLEV( L ) .GT. PRES( M ) ) THEN

                            !  found plev(L)

                            LEVS( C,R,L ) = M-1

                            CLIN( C,R,L ) =  
     &                           ( PLEV(L)   - PRES(M) )/ 
     &                           ( PRES(M-1) - PRES(M) )
     &                                     

                            CLOG( C,R,L ) =   
     &                      ( LOG( PLEV(L) ) - PLOG(M) )/ 
     &                           ( PLOG(M-1) - PLOG(M) )

                            CPOW( C,R,L ) = 
     &                       ( PLEV(L)**ROCP - PPOW(M) )/ 
     &                           ( PPOW(M-1) - PPOW(M) )

                            N = M-1
                            GO TO  11       !  done with this (C,R,L)

                        END IF

                    END DO          !  end loop on remaining MM5 levels

                    !  if you get to here:  never found PLEV(L); error

                    CALL M3EXIT( 'MCPL:MMD2P', 0, 0, 
     &                           'PLEV not found', 2 )

11                  CONTINUE        !  target of "found plev(L)"

                END DO  !  end loop on output layers L

22          CONTINUE    !  end loop on output cols C
33      CONTINUE        !  end loop on output rows R

        RETURN
        END

